@@ -1,3 +1,4 @@
+/* SPDX-License-Identifier: GPL-2.0-only */
 /* can-calc-bit-timing.c: Calculate CAN bit timing parameters
  *
  * Copyright (C) 2008 Wolfgang Grandegger <wg@grandegger.com>
@@ -11,13 +12,13 @@
  *   email:pisa@cmp.felk.cvut.cz
  *
  * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * any later version.
+ * it under the terms of the version 2 of the GNU General Public License
+ * as published by the Free Software Foundation
  */
 
 #include <errno.h>
 #include <getopt.h>
+#include <libgen.h>
 #include <limits.h>
 #include <stdbool.h>
 #include <stdint.h>
@@ -25,8 +26,8 @@
 #include <stdlib.h>
 #include <string.h>
 
-#include <linux/types.h>
 #include <linux/can/netlink.h>
+#include <linux/types.h>
 
 /* imported from kernel */
 
@@ -131,17 +132,16 @@ static inline void *netdev_priv(const struct net_device *dev)
 
 static void print_usage(char *cmd)
 {
+	printf("%s - calculate CAN bit timing parameters.\n", cmd);
 	printf("Usage: %s [options] [<CAN-contoller-name>]\n"
-	       "\tOptions:\n"
-	       "\t-q           : don't print header line\n"
-	       "\t-l           : list all support CAN controller names\n"
-	       "\t-b <bitrate> : bit-rate in bits/sec\n"
-	       "\t-s <samp_pt> : sample-point in one-tenth of a percent\n"
+	       "Options:\n"
+	       "\t-q             don't print header line\n"
+	       "\t-l             list all support CAN controller names\n"
+	       "\t-b <bitrate>   bit-rate in bits/sec\n"
+	       "\t-s <samp_pt>   sample-point in one-tenth of a percent\n"
 	       "\t               or 0 for CIA recommended sample points\n"
-	       "\t-c <clock>   : real CAN system clock in Hz\n",
+	       "\t-c <clock>     real CAN system clock in Hz\n",
 	       cmd);
-
-	exit(EXIT_FAILURE);
 }
 
 static void printf_btr_sja1000(struct can_bittiming *bt, bool hdr)
@@ -444,7 +444,7 @@ static int can_calc_bittiming(struct net_device *dev, struct can_bittiming *bt,
 	unsigned int spt_nominal;	/* nominal sample point */
 	unsigned int best_tseg = 0;	/* current best value for tseg */
 	unsigned int best_brp = 0;	/* current best value for brp */
-	unsigned int brp, tsegall, tseg, tseg1, tseg2;
+	unsigned int brp, tsegall, tseg, tseg1 = 0, tseg2 = 0;
 	u64 v64;
 
 	/* Use CiA recommended sample points */
@@ -638,7 +638,7 @@ int main(int argc, char *argv[])
 
 	const struct calc_bittiming_const *btc;
 
-	while ((opt = getopt(argc, argv, "b:c:lqs:")) != -1) {
+	while ((opt = getopt(argc, argv, "b:c:lqs:?")) != -1) {
 		switch (opt) {
 		case 'b':
 			bitrate_nominal = atoi(optarg);
@@ -660,14 +660,22 @@ int main(int argc, char *argv[])
 			spt_nominal = strtoul(optarg, NULL, 10);
 			break;
 
+		case '?':
+			print_usage(basename(argv[0]));
+			exit(EXIT_SUCCESS);
+			break;
+
 		default:
-			print_usage(argv[0]);
+			print_usage(basename(argv[0]));
+			exit(EXIT_FAILURE);
 			break;
 		}
 	}
 
-	if (argc > optind + 1)
+	if (argc > optind + 1) {
 		print_usage(argv[0]);
+		exit(EXIT_FAILURE);
+	}
 
 	if (argc == optind + 1)
 		name = argv[optind];
@@ -677,11 +685,14 @@ int main(int argc, char *argv[])
 		exit(EXIT_SUCCESS);
 	}
 
-	if (spt_nominal && (spt_nominal >= 1000 || spt_nominal < 100))
+	if (spt_nominal && (spt_nominal >= 1000 || spt_nominal < 100)) {
 		print_usage(argv[0]);
+		exit(EXIT_FAILURE);
+	}
 
 	for (i = 0; i < ARRAY_SIZE(can_calc_consts); i++) {
-		if (name && strcmp(can_calc_consts[i].bittiming_const.name, name))
+		if (name &&
+		    strcmp(can_calc_consts[i].bittiming_const.name, name) != 0)
 			continue;
 
 		found = true;
