@@ -1,3 +1,4 @@
+/* SPDX-License-Identifier: (GPL-2.0-only OR BSD-3-Clause) */
 /*
  * cangw.c - manage PF_CAN netlink gateway
  *
@@ -41,17 +42,17 @@
  *
  */
 
-#include <stdio.h>
-#include <stdlib.h>
-#include <libgen.h>
-#include <string.h>
-#include <unistd.h>
 #include <errno.h>
-#include <sys/socket.h>
-#include <net/if.h>
+#include <libgen.h>
+#include <linux/can/gw.h>
 #include <linux/netlink.h>
 #include <linux/rtnetlink.h>
-#include <linux/can/gw.h>
+#include <net/if.h>
+#include <stdio.h>
+#include <stdlib.h>
+#include <string.h>
+#include <sys/socket.h>
+#include <unistd.h>
 
 enum {
 	UNSPEC,
@@ -67,13 +68,19 @@ struct modattr {
 	__u8 instruction;
 } __attribute__((packed));
 
+struct fdmodattr {
+	struct canfd_frame cf;
+	__u8 modtype;
+	__u8 instruction;
+} __attribute__((packed));
+
 
 #define RTCAN_RTA(r)  ((struct rtattr*)(((char*)(r)) + NLMSG_ALIGN(sizeof(struct rtcanmsg))))
 #define RTCAN_PAYLOAD(n) NLMSG_PAYLOAD(n,sizeof(struct rtcanmsg))
 
 /* some netlink helpers stolen from iproute2 package */
 #define NLMSG_TAIL(nmsg) \
-        ((struct rtattr *)(((void *) (nmsg)) + NLMSG_ALIGN((nmsg)->nlmsg_len)))
+        ((struct rtattr *)(((char *) (nmsg)) + NLMSG_ALIGN((nmsg)->nlmsg_len)))
 
 int addattr_l(struct nlmsghdr *n, int maxlen, int type, const void *data,
 	      int alen)
@@ -81,7 +88,7 @@ int addattr_l(struct nlmsghdr *n, int maxlen, int type, const void *data,
 	int len = RTA_LENGTH(alen);
 	struct rtattr *rta;
 
-	if (NLMSG_ALIGN(n->nlmsg_len) + RTA_ALIGN(len) > maxlen) {
+	if ((int)(NLMSG_ALIGN(n->nlmsg_len) + RTA_ALIGN(len)) > maxlen) {
 		fprintf(stderr, "addattr_l: message exceeded bound of %d\n",
 			maxlen);
 		return -1;
@@ -124,7 +131,36 @@ void printmod(const char *type, const void *data)
 
 	printf(":%03X.%X.", mod.cf.can_id, mod.cf.can_dlc);
 
-	for (i = 0; i < 8; i++)
+	for (i = 0; i < CAN_MAX_DLEN; i++)
+		printf("%02X", mod.cf.data[i]);
+
+	printf(" ");
+}
+
+void printfdmod(const char *type, const void *data)
+{
+	struct fdmodattr mod;
+	int i;
+
+	memcpy (&mod, data, CGW_FDMODATTR_LEN);
+
+	printf("-M %s:", type);
+
+	if (mod.modtype & CGW_MOD_ID)
+		printf("I");
+
+	if (mod.modtype & CGW_MOD_FLAGS)
+		printf("F");
+
+	if (mod.modtype & CGW_MOD_LEN)
+		printf("L");
+
+	if (mod.modtype & CGW_MOD_DATA)
+		printf("D");
+
+	printf(":%03X.%X.%X.", mod.cf.can_id, mod.cf.flags, mod.cf.len);
+
+	for (i = 0; i < CANFD_MAX_DLEN; i++)
 		printf("%02X", mod.cf.data[i]);
 
 	printf(" ");
@@ -186,45 +222,60 @@ void print_cs_crc8(struct cgw_csum_crc8 *cs_crc8)
 
 void print_usage(char *prg)
 {
+	fprintf(stderr, "%s - manage PF_CAN netlink gateway.\n", prg);
 	fprintf(stderr, "\nUsage: %s [options]\n\n", prg);
-	fprintf(stderr, "Commands:  -A (add a new rule)\n");
-	fprintf(stderr, "           -D (delete a rule)\n");
-	fprintf(stderr, "           -F (flush / delete all rules)\n");
-	fprintf(stderr, "           -L (list all rules)\n");
-	fprintf(stderr, "Mandatory: -s <src_dev>  (source netdevice)\n");
-	fprintf(stderr, "           -d <dst_dev>  (destination netdevice)\n");
-	fprintf(stderr, "Options:   -t (preserve src_dev rx timestamp)\n");
-	fprintf(stderr, "           -e (echo sent frames - recommended on vcanx)\n");
-	fprintf(stderr, "           -i (allow to route to incoming interface)\n");
-	fprintf(stderr, "           -u <uid> (user defined modification identifier)\n");
-	fprintf(stderr, "           -l <hops> (limit the number of frame hops / routings)\n");
-	fprintf(stderr, "           -f <filter> (set CAN filter)\n");
-	fprintf(stderr, "           -m <mod> (set frame modifications)\n");
-	fprintf(stderr, "           -x <from_idx>:<to_idx>:<result_idx>:<init_xor_val> (XOR checksum)\n");
-	fprintf(stderr, "           -c <from>:<to>:<result>:<init_val>:<xor_val>:<crctab[256]> (CRC8 cs)\n");
-	fprintf(stderr, "           -p <profile>:[<profile_data>] (CRC8 checksum profile & parameters)\n");
+	fprintf(stderr, "Commands:\n");
+	fprintf(stderr, "          -A  (add a new rule)\n");
+	fprintf(stderr, "          -D  (delete a rule)\n");
+	fprintf(stderr, "          -F  (flush / delete all rules)\n");
+	fprintf(stderr, "          -L  (list all rules)\n");
+	fprintf(stderr, "Mandatory:\n");
+	fprintf(stderr, "          -s <src_dev>  (source netdevice)\n");
+	fprintf(stderr, "          -d <dst_dev>  (destination netdevice)\n");
+	fprintf(stderr, "Options:\n");
+	fprintf(stderr, "          -X  (this is a CAN FD rule)\n");
+	fprintf(stderr, "          -t  (preserve src_dev rx timestamp)\n");
+	fprintf(stderr, "          -e  (echo sent frames - recommended on vcanx)\n");
+	fprintf(stderr, "          -i  (allow to route to incoming interface)\n");
+	fprintf(stderr, "          -u <uid>  (user defined modification identifier)\n");
+	fprintf(stderr, "          -l <hops>  (limit the number of frame hops / routings)\n");
+	fprintf(stderr, "          -f <filter>  (set CAN filter)\n");
+	fprintf(stderr, "          -m <mod>  (set Classic CAN frame modifications)\n");
+	fprintf(stderr, "          -M <MOD>  (set CAN FD frame modifications)\n");
+	fprintf(stderr, "          -x <from_idx>:<to_idx>:<result_idx>:<init_xor_val>  (XOR checksum)\n");
+	fprintf(stderr, "          -c <from>:<to>:<result>:<init_val>:<xor_val>:<crctab[256]>  (CRC8 cs)\n");
+	fprintf(stderr, "          -p <profile>:[<profile_data>]  (CRC8 checksum profile & parameters)\n");
 	fprintf(stderr, "\nValues are given and expected in hexadecimal values. Leading 0s can be omitted.\n");
 	fprintf(stderr, "\n");
-	fprintf(stderr, "<filter> is a <value><mask> CAN identifier filter\n");
-	fprintf(stderr, "   <can_id>:<can_mask> (matches when <received_can_id> & mask == can_id & mask)\n");
-	fprintf(stderr, "   <can_id>~<can_mask> (matches when <received_can_id> & mask != can_id & mask)\n");
+	fprintf(stderr, "<filter> is a <value><mask> CAN identifier filter:\n");
+	fprintf(stderr, "  <can_id>:<can_mask>  (matches when <received_can_id> & mask == can_id & mask)\n");
+	fprintf(stderr, "  <can_id>~<can_mask>  (matches when <received_can_id> & mask != can_id & mask)\n");
 	fprintf(stderr, "\n");
 	fprintf(stderr, "<mod> is a CAN frame modification instruction consisting of\n");
 	fprintf(stderr, "<instruction>:<can_frame-elements>:<can_id>.<can_dlc>.<can_data>\n");
-	fprintf(stderr, " - <instruction> is one of 'AND' 'OR' 'XOR' 'SET'\n");
-	fprintf(stderr, " - <can_frame-elements> is _one_ or _more_ of 'I'dentifier 'L'ength 'D'ata\n");
-	fprintf(stderr, " - <can_id> is an u32 value containing the CAN Identifier\n");
-	fprintf(stderr, " - <can_dlc> is an u8 value containing the data length code (0 .. 8)\n");
-	fprintf(stderr, " - <can_data> is always eight(!) u8 values containing the CAN frames data\n");
+	fprintf(stderr, "  <instruction>  is one of 'AND' 'OR' 'XOR' 'SET'\n");
+	fprintf(stderr, "  <can_frame-elements>  is _one_ or _more_ of 'I'dentifier 'L'ength 'D'ata\n");
+	fprintf(stderr, "  <can_id>  is an u32 value containing the CAN Identifier\n");
+	fprintf(stderr, "  <can_dlc>  is an u8 value containing the data length code (0 .. 8)\n");
+	fprintf(stderr, "  <can_data>  is always eight(!) u8 values containing the CAN frames data\n");
+	fprintf(stderr, "\n");
+	fprintf(stderr, "<MOD> is a CAN FD frame modification instruction consisting of\n");
+	fprintf(stderr, "<instruction>:<canfd_frame-elements>:<can_id>.<flags>.<len>.<can_data>\n");
+	fprintf(stderr, "  <instruction>  is one of 'AND' 'OR' 'XOR' 'SET'\n");
+	fprintf(stderr, "  <canfd_frame-elements>  is _one_ or _more_ of 'I'd 'F'lags 'L'ength 'D'ata\n");
+	fprintf(stderr, "  <can_id>  is an u32 value containing the CAN FD Identifier\n");
+	fprintf(stderr, "  <flags>  is an u8 value containing CAN FD flags (CANFD_BRS, CANFD_ESI)\n");
+	fprintf(stderr, "  <len>  is an u8 value containing the data length (0 .. 64)\n");
+	fprintf(stderr, "  <can_data>  is always 64(!) u8 values containing the CAN FD frames data\n");
 	fprintf(stderr, "The max. four modifications are performed in the order AND -> OR -> XOR -> SET\n");
 	fprintf(stderr, "\n");
-	fprintf(stderr, "Example:\n");
-	fprintf(stderr, "%s -A -s can0 -d vcan3 -e -f 123:C00007FF -m SET:IL:333.4.1122334455667788\n", prg);
-	fprintf(stderr, "\n");
 	fprintf(stderr, "Supported CRC 8 profiles:\n");
-	fprintf(stderr, "Profile '%d' (1U8)       - add one additional u8 value\n", CGW_CRC8PRF_1U8);
-	fprintf(stderr, "Profile '%d' (16U8)      - add u8 value from table[16] indexed by (data[1] & 0xF)\n", CGW_CRC8PRF_16U8);
-	fprintf(stderr, "Profile '%d' (SFFID_XOR) - add u8 value (can_id & 0xFF) ^ (can_id >> 8 & 0xFF)\n", CGW_CRC8PRF_SFFID_XOR);
+	fprintf(stderr, " Profile '%d' (1U8)        add one additional u8 value\n", CGW_CRC8PRF_1U8);
+	fprintf(stderr, " Profile '%d' (16U8)       add u8 value from table[16] indexed by (data[1] & 0xF)\n", CGW_CRC8PRF_16U8);
+	fprintf(stderr, " Profile '%d' (SFFID_XOR)  add u8 value (can_id & 0xFF) ^ (can_id >> 8 & 0xFF)\n", CGW_CRC8PRF_SFFID_XOR);
+	fprintf(stderr, "\n");
+	fprintf(stderr, "Examples:\n");
+	fprintf(stderr, "%s -A -s can0 -d vcan3 -e -f 123:C00007FF -m SET:IL:333.4.1122334455667788\n", prg);
 	fprintf(stderr, "\n");
 }
 
@@ -244,14 +295,14 @@ int parse_crc8_profile(char *optarg, struct cgw_csum_crc8 *crc8)
 	int ret = 1;
 	char *ptr;
 
-	if (sscanf(optarg, "%hhd:", &crc8->profile) != 1)
+	if (sscanf(optarg, "%hhu:", &crc8->profile) != 1)
 		return ret;
 
 	switch (crc8->profile) {
 
 	case  CGW_CRC8PRF_1U8:
 
-		if (sscanf(optarg, "%hhd:%2hhx", &crc8->profile, &crc8->profile_data[0]) == 2)
+		if (sscanf(optarg, "%hhu:%2hhx", &crc8->profile, &crc8->profile_data[0]) == 2)
 			ret = 0;
 
 		break;
@@ -281,7 +332,8 @@ int parse_crc8_profile(char *optarg, struct cgw_csum_crc8 *crc8)
 int parse_mod(char *optarg, struct modattr *modmsg)
 {
 	char *ptr, *nptr;
-	char hexdata[17] = {0};
+	char hexdata[(CAN_MAX_DLEN * 2) + 1] = {0};
+	canid_t can_id;
 
 	ptr = optarg;
 	nptr = strchr(ptr, ':');
@@ -330,30 +382,97 @@ int parse_mod(char *optarg, struct modattr *modmsg)
 		ptr++;
 	}
 
-	if (sscanf(++ptr, "%x.%hhx.%16s", &modmsg->cf.can_id,
+	if (sscanf(++ptr, "%x.%hhx.%16s", &can_id,
 		   (unsigned char *)&modmsg->cf.can_dlc, hexdata) != 3)
 		return 5;
 
-	/* 4-bit masks can have values from 0 to 0xF */ 
-	if (modmsg->cf.can_dlc > 0xF)
+	modmsg->cf.can_id = can_id;
+
+	if (strlen(hexdata) != (CAN_MAX_DLEN * 2))
 		return 6;
 
-	/* but when setting CAN_DLC the value has to be limited to 8 */
-	if (modmsg->instruction == CGW_MOD_SET && modmsg->cf.can_dlc > 8)
+	if (b64hex(hexdata, &modmsg->cf.data[0], CAN_MAX_DLEN))
 		return 7;
 
-	if (strlen(hexdata) != 16)
-		return 8;
+	return 0; /* ok */
+}
 
-	if (b64hex(hexdata, &modmsg->cf.data[0], 8))
-		return 9;
+int parse_fdmod(char *optarg, struct fdmodattr *modmsg)
+{
+	char *ptr, *nptr;
+	char hexdata[(CANFD_MAX_DLEN * 2) + 1] = {0};
+	canid_t can_id;
+
+	ptr = optarg;
+	nptr = strchr(ptr, ':');
+
+	if ((nptr - ptr > 3) || (nptr - ptr == 0))
+		return 1;
+
+	if (!strncmp(ptr, "AND", 3))
+		modmsg->instruction = CGW_FDMOD_AND;
+	else if (!strncmp(ptr, "OR", 2))
+		modmsg->instruction = CGW_FDMOD_OR;
+	else if (!strncmp(ptr, "XOR", 3))
+		modmsg->instruction = CGW_FDMOD_XOR;
+	else if (!strncmp(ptr, "SET", 3))
+		modmsg->instruction = CGW_FDMOD_SET;
+	else
+		return 2;
+
+	ptr = nptr+1;
+	nptr = strchr(ptr, ':');
+
+	if ((nptr - ptr > 4) || (nptr - ptr == 0))
+		return 3;
+
+	modmsg->modtype = 0;
+
+	while (*ptr != ':') {
+
+		switch (*ptr) {
+
+		case 'I':
+			modmsg->modtype |= CGW_MOD_ID;
+			break;
+
+		case 'F':
+			modmsg->modtype |= CGW_MOD_FLAGS;
+			break;
+
+		case 'L':
+			modmsg->modtype |= CGW_MOD_LEN;
+			break;
+
+		case 'D':
+			modmsg->modtype |= CGW_MOD_DATA;
+			break;
+
+		default:
+			return 4;
+		}
+		ptr++;
+	}
+
+	if (sscanf(++ptr, "%x.%hhx.%hhx.%128s", &can_id,
+		   (unsigned char *)&modmsg->cf.flags,
+		   (unsigned char *)&modmsg->cf.len, hexdata) != 4)
+		return 5;
+
+	modmsg->cf.can_id = can_id;
+
+	if (strlen(hexdata) != (CANFD_MAX_DLEN * 2))
+		return 6;
+
+	if (b64hex(hexdata, &modmsg->cf.data[0], CANFD_MAX_DLEN))
+		return 7;
 
 	return 0; /* ok */
 }
 
 int parse_rtlist(char *prgname, unsigned char *rxbuf, int len)
 {
-	char ifname[IF_NAMESIZE]; /* internface name for if_indextoname() */
+	char ifname[IF_NAMESIZE]; /* interface name for if_indextoname() */
 	struct rtcanmsg *rtc;
 	struct rtattr *rta;
 	struct nlmsghdr *nlh;
@@ -419,6 +538,10 @@ int parse_rtlist(char *prgname, unsigned char *rxbuf, int len)
 			case CGW_MOD_OR:
 			case CGW_MOD_XOR:
 			case CGW_MOD_SET:
+			case CGW_FDMOD_AND:
+			case CGW_FDMOD_OR:
+			case CGW_FDMOD_XOR:
+			case CGW_FDMOD_SET:
 			case CGW_MOD_UID:
 			case CGW_LIM_HOPS:
 			case CGW_CS_XOR:
@@ -456,6 +579,9 @@ int parse_rtlist(char *prgname, unsigned char *rxbuf, int len)
 		printf("-s %s ", if_indextoname(src_ifindex, ifname));
 		printf("-d %s ", if_indextoname(dst_ifindex, ifname));
 
+		if (rtc->flags & CGW_FLAGS_CAN_FD)
+			printf("-X ");
+
 		if (rtc->flags & CGW_FLAGS_CAN_ECHO)
 			printf("-e ");
 
@@ -493,6 +619,22 @@ int parse_rtlist(char *prgname, unsigned char *rxbuf, int len)
 				printmod("SET", RTA_DATA(rta));
 				break;
 
+			case CGW_FDMOD_AND:
+				printfdmod("AND", RTA_DATA(rta));
+				break;
+
+			case CGW_FDMOD_OR:
+				printfdmod("OR", RTA_DATA(rta));
+				break;
+
+			case CGW_FDMOD_XOR:
+				printfdmod("XOR", RTA_DATA(rta));
+				break;
+
+			case CGW_FDMOD_SET:
+				printfdmod("SET", RTA_DATA(rta));
+				break;
+
 			case CGW_MOD_UID:
 				printf("-u %X ", *(__u32 *)RTA_DATA(rta));
 				break;
@@ -548,8 +690,7 @@ int main(int argc, char **argv)
 	struct {
 		struct nlmsghdr nh;
 		struct rtcanmsg rtcan;
-		char buf[600];
-
+		char buf[1500];
 	} req;
 
 	unsigned char rxbuf[8192]; /* netlink receive buffer */
@@ -565,19 +706,19 @@ int main(int argc, char **argv)
 	struct can_filter filter;
 	struct sockaddr_nl nladdr;
 
-	struct cgw_csum_xor cs_xor;
-	struct cgw_csum_crc8 cs_crc8;
+	struct cgw_csum_xor cs_xor = { 0 };
+	struct cgw_csum_crc8 cs_crc8 = { 0 };
 	char crc8tab[513] = {0};
 
 	struct modattr modmsg[CGW_MOD_FUNCS];
+	struct fdmodattr fdmodmsg[CGW_MOD_FUNCS];
 	int modidx = 0;
+	int fdmodidx = 0;
 	int i;
 
 	memset(&req, 0, sizeof(req));
-	memset(&cs_xor, 0, sizeof(cs_xor));
-	memset(&cs_crc8, 0, sizeof(cs_crc8));
 
-	while ((opt = getopt(argc, argv, "ADFLs:d:teiu:l:f:c:p:x:m:?")) != -1) {
+	while ((opt = getopt(argc, argv, "ADFLs:d:Xteiu:l:f:c:p:x:m:M:?")) != -1) {
 		switch (opt) {
 
 		case 'A':
@@ -608,6 +749,10 @@ int main(int argc, char **argv)
 			dst_ifindex = if_nametoindex(optarg);
 			break;
 
+		case 'X':
+			flags |= CGW_FLAGS_CAN_FD;
+			break;
+
 		case 't':
 			flags |= CGW_FLAGS_CAN_SRC_TSTAMP;
 			break;
@@ -625,7 +770,7 @@ int main(int argc, char **argv)
 			break;
 
 		case 'l':
-			if (sscanf(optarg, "%hhd", &limit_hops) != 1 || !(limit_hops)) {
+			if (sscanf(optarg, "%hhu", &limit_hops) != 1 || !(limit_hops)) {
 				printf("Bad hop limit definition '%s'.\n", optarg);
 				exit(1);
 			}
@@ -685,6 +830,14 @@ int main(int argc, char **argv)
 			}
 			break;
 
+		case 'M':
+			/* may be triggered by each of the CGW_FDMOD_FUNCS functions */
+			if ((fdmodidx < CGW_MOD_FUNCS) && (err = parse_fdmod(optarg, &fdmodmsg[fdmodidx++]))) {
+				printf("Problem %d with modification definition '%s'.\n", err, optarg);
+				exit(1);
+			}
+			break;
+
 		case '?':
 			print_usage(basename(argv[0]));
 			exit(0);
@@ -709,8 +862,20 @@ int main(int argc, char **argv)
 		exit(1);
 	}
 
-	if (!modidx && (have_cs_crc8 || have_cs_xor)) {
-		printf("-c or -x can only be used in conjunction with -m\n");
+	if (flags & CGW_FLAGS_CAN_FD) {
+		if (modidx) {
+			printf("No -m modifications allowed in CAN FD mode!\n");
+			exit(1);
+		}
+	} else {
+		if (fdmodidx) {
+			printf("No -M modifications allowed in Classic CAN mode!\n");
+			exit(1);
+		}
+	}
+
+	if ((!modidx && !fdmodidx) && (have_cs_crc8 || have_cs_xor)) {
+		printf("-c or -x can only be used in conjunction with -m/-M\n");
 		exit(1);
 	}
 
@@ -784,6 +949,10 @@ int main(int argc, char **argv)
 	for (i = 0; i < modidx; i++)
 		addattr_l(&req.nh, sizeof(req), modmsg[i].instruction, &modmsg[i], CGW_MODATTR_LEN);
 
+	/* add up to CGW_FDMOD_FUNCS modification definitions */
+	for (i = 0; i < fdmodidx; i++)
+		addattr_l(&req.nh, sizeof(req), fdmodmsg[i].instruction, &fdmodmsg[i], CGW_FDMODATTR_LEN);
+
 	memset(&nladdr, 0, sizeof(nladdr));
 	nladdr.nl_family = AF_NETLINK;
 	nladdr.nl_pid    = 0;
